from rest import decorators as rd
from rest import crypto
from rest.mail import render_to_mail
from rest import views as rv
# from rest.views import restStatus, restGet, restPermissionDenied
from rest.jwtoken import JWToken, JWT_KEY
from rest import helpers
from rest import settings
from account import models as am
from medialib.qrcode import generateQRCode
from django.http import HttpResponse
from datetime import datetime, timedelta

ALLOW_BASIC_LOGIN = settings.get("ALLOW_BASIC_LOGIN", False)
FORGET_ALWAYS_TRUE = settings.get("FORGET_ALWAYS_TRUE", True)
JWT_UPDATE_REFRESH_TOKEN = settings.get("JWT_UPDATE_REFRESH_TOKEN", False)


@rd.urlPOST(r'^login$')
@rd.urlPOST(r'^login/$')
@rd.never_cache
def member_login(request):
    if not ALLOW_BASIC_LOGIN:
        return jwt_login(request)
    username = request.DATA.get('username', None)
    auth_code = request.DATA.get(["auth_code", "code", "invite_token"], None)
    if username and auth_code:
        return member_login_uname_code(request, username, auth_code)
    password = request.DATA.get('password', None)
    if username and password:
        return member_login_uname_pword(request, username, password)
    return rv.restPermissionDenied(request, f"Invalid credentials {username}/{auth_code}", 401)


@rd.urlPOST(r'^jwt/login$')
@rd.urlPOST(r'^jwt/login/$')
@rd.never_cache
def jwt_login(request):
    # poor mans JWT, carried over
    username = request.DATA.get('username', None)
    if not username:
        return rv.restPermissionDenied(request, "Password and/or Username is incorrect", error_code=422)
    member = getMemberByUsername(username)
    if not member:
        return rv.restPermissionDenied(request, error=f"Password and/or Username is incorrect for {username}", error_code=422)
    auth_code = request.DATA.get(["auth_code", "code", "invite_token"], None)
    if username and auth_code:
        return member_login_uname_code(request, username, auth_code)
    password = request.DATA.get('password', None)
    member.canLogin(request)  # throws exception if cannot login
    if member.requires_totp or member.has_totp:
        resp = checkForTOTP(request, member)
        if resp is not None:
            return resp
    if not member.login(request=request, password=password, use_jwt=True):
        # we do not want permission denied catcher invoked as it is already handled in login method
        return rv.restStatus(request, False, error=f"Invalid Credentials {username}", error_code=401)
    return on_complete_jwt(request, member)


def on_complete_jwt(request, member):
    if member.security_token is None or member.security_token == JWT_KEY or member.force_single_session:
        member.refreshSecurityToken()

    member.log(
        "jwt_login", "jwt login succesful", 
        request, method="login", level=7)

    device_id = request.DATA.get(["device_id", "deviceID"])
    token = JWToken(user_id=member.pk, key=member.security_token, device_id=device_id)

    request.user = member.getUser()
    request.member = member
    request.signature = token.session_id
    request.device_id = device_id
    request.buid = request.DATA.get("__buid__", None)
    request.auth_session = am.AuthSession.NewSession(request)
    if bool(device_id):
        am.MemberDevice.register(request, member, device_id)
    
    request.jwt_token = token.access_token  # this tells the middleware to store in cookie
    return rv.restGet(
        request, 
        dict(
            access=token.access_token,
            refresh=token.refresh_token,
            id=member.pk))


@rd.urlPOST(r'^jwt/logout$')
@rd.urlPOST(r'^jwt/logout/$')
@rd.never_cache
def jwt_logout(request):
    # this will force our token to change
    if request.member:
        request.member.log("jwt_logout", "jwt logout", request, method="logout", level=25)
        request.member.sendEvent("logout", "user requested logout")
        request.member.refreshSecurityToken()
        request.clear_jwt_cookie = True  # tells middleware to remove from cookie
    return rv.restStatus(request, True)


@rd.urlPOST(r'^jwt/refresh$')
@rd.urlPOST(r'^jwt/refresh/$')
@rd.never_cache
def jwt_refresh(request):
    # poor mans JWT, carried over
    rtoken = request.DATA.get(['refresh_token', 'refresh'], None)
    if not bool(rtoken):
        return rv.restPermissionDenied(request, error="requires token", error_code=703)
    token = JWToken(token=rtoken)
    if not token.payload.user_id:
        return rv.restPermissionDenied(request, error="invalid token", error_code=-701)
    member = am.Member.objects.filter(pk=token.payload.user_id).last()
    if not member:
        return rv.restPermissionDenied(request, error=f"Password or Username is incorrect for uid:{token.payload.user_id}", error_code=422)
    token.key = member.security_token
    if not token.is_valid:
        request.member = member
        return rv.restPermissionDenied(request, error="invalid token", error_code=-702)
    if member.security_token is None:
        member.refreshSecurityToken()
    member.canLogin(request)
    token.refresh()
    request.jwt_token = token.access_token  # this tells the middleware to store in cookie
    if JWT_UPDATE_REFRESH_TOKEN:
        rtoken = token.refresh_token
    return rv.restGet(request, dict(access=token.access_token, refresh=rtoken))


def getMemberByUsername(username):
    member = None
    username = username.lower()
    if username.count('@') == 1:
        member = am.Member.objects.filter(email=username).last()
    if not member:
        member = am.Member.objects.filter(username=username).last()
    return member


def checkForTOTP(request, member):
    if not member.has_totp and not member.phone_number:
        member.reportIncident(
            "account", f"{member.username} TOTP not set", level=8,
            error_code=455,
            request=request)
        return None
    if not member.has_totp:
        # we have a phone number so tell them to login with code
        # they will need to request a code
        request.member = member
        return rv.restPermissionDenied(
            request, error=member.phone_number[-4:],
            error_code=454)
    totp_code = request.DATA.get("totp_code", None)
    if totp_code is None:
        # member.log("login_blocked", "requires MFA (TOTP)", request, method="login", level=31)
        request.member = member
        return rv.restPermissionDenied(request, error="Requires MFA (TOTP)", error_code=455)
    if not member.totp_verify(totp_code):
        request.member = member
        member.log("login_blocked", "Invalid MFA code", request, method="login", level=31)
        return rv.restPermissionDenied(request, error="Invalid Credentials", error_code=456)
    return None


def member_login_uname_pword(request, username, password):
    member = getMemberByUsername(username)
    if not member:
        return rv.restPermissionDenied(request, error=f"Password or Username is not correct for {username}", error_code=422)
    member.canLogin(request)  # throws exception if cannot login
    if member.requires_totp or member.has_totp:
        resp = checkForTOTP(request, member)
        if resp is not None:
            return resp
    if not member.login(request=request, password=password, use_jwt=False):
        request.member = member
        member.log("login_failed", "incorrect password", request, method="login", level=31)
        return rv.restPermissionDenied(request, error="Password or Username is incorrect", error_code=401)

    member.log("password_login", "password login", request, method="login", level=7)
    if request.session is not None:
        request.session["member_id"] = member.pk
        request.session["_auth_user_id"] = member.pk
    return rv.restGet(request, dict(id=member.pk, session_key=request.session.session_key))


def member_login_uname_code(request, username, auth_code):
    member = getMemberByUsername(username)
    if not member:
        return rv.restPermissionDenied(request, error=f"Username or code is incorrect {username}/{auth_code}", error_code=422)
    if not member.is_active:
        request.member = member
        member.log("login_blocked", "account is not active", request, method="login", level=31)
        return rv.restPermissionDenied(request, error=f"{username} Account disabled", error_code=410)
    if member.is_blocked:
        request.member = member
        member.log("login_blocked", "account is locked out", request, method="login", level=31)
        return rv.restPermissionDenied(request, error=f"{username} Account locked out", error_code=411)
    auth_code = auth_code.replace('-', '').replace(' ', '')
    if member.auth_code is None or member.auth_code != auth_code:
        request.member = member
        member.log("login_blocked", "accessing used or expired token link", request, method="login", level=31)
        return rv.restPermissionDenied(request, f"token already used or expired for {username}", error_code=492)
    if member.auth_code_expires < datetime.now():
        request.member = member
        member.log("login_blocked", "accessing expired token link", request, method="login", level=31)
        return rv.restPermissionDenied(request, "token expired", error_code=493)
    password = request.DATA.get('new_password', None)
    if password:
        member.setPassword(password)
    member.auth_code = None
    member.auth_code_expires = None
    member.login(request=request)
    member.save()
    member.log("code_login", "code login", request, method="login", level=8)
    if request.DATA.get("auth_method") == "basic":
        return rv.restGet(request, dict(id=member.pk, session_key=request.session.session_key))
    return on_complete_jwt(request, member)


@rd.url(r'^logout$')
@rd.url(r'^logout/$')
@rd.never_cache
def member_logout(request):
    """
    | Parameters: none

    | Return: status + error

    | Logout
    """
    if request.user.is_authenticated:
        request.user.log("logout", "user logged out", request, method="logout", level=8)
    request.member.logout(request)
    return rv.restStatus(request, True)


@rd.url(r'^loggedin/$')
@rd.never_cache
def is_member_logged_in(request):
    """
    | param: none

    | Return: status + error

    | Check if the current user is logged in
    """
    if request.user:
        return rv.restStatus(request, request.user.is_authenticated)
    return rv.restStatus(request, False)


@rd.urlPOST('invite/validate')
@rd.requires_params(["username"])
def member_invite_confirm(request):
    username = request.DATA.get('username', None)
    auth_code = request.DATA.get(["auth_token", "invite_token"], None)
    member = getMemberByUsername(username)
    if not member or not auth_code:
        return rv.restPermissionDenied(request, error=f"Username or code is incorrect {username}/{auth_code}", error_code=422)
    auth_code = auth_code.replace('-', '').replace(' ', '')
    if member.auth_code is None or member.auth_code != auth_code:
        return rv.restStatus(request, False)
    return rv.restStatus(request, True)


@rd.urlPOST('mfa/request_code')
@rd.requires_params(["username"])
def member_request_code(request):
    member, resp = get_member_from_request(request)
    if resp is not None:
        return resp
    return member_forgot_password_code(request, member)


def get_member_from_request(request):
    username = request.DATA.get('username', None)
    if not username:
        return None, rv.restPermissionDenied("Username is required")
    member = getMemberByUsername(username)
    if not member:
        return member, rv.restPermissionDenied(request, error=f"Password or Username is incorrect for {username}", error_code=422)
    if not member.is_active:
        request.member = member
        member.log("login_blocked", "account is not active", request, method="login", level=31)
        return member, rv.restPermissionDenied(request, error=f"{member.username} Account disabled", error_code=410)
    if member.is_blocked:
        request.member = member
        member.log("login_blocked", "account is locked out", request, method="login", level=31)
        return member, rv.restPermissionDenied(request, error=f"{member.username} Account locked out", error_code=411)
    return member, None  


@rd.urlPOST('forgot')
@rd.never_cache
def member_forgot_password(request):
    """
    | param: username = use the username as the lookup
    | param: email = use the email as the lookup

    | Return: status + error

    | Send fgroupet password reset instructions
    """
    member, resp = get_member_from_request(request)
    if member is None:
        if FORGET_ALWAYS_TRUE:
            return rv.restStatus(request, True, msg="Password reset instructions have been sent to your email. (If valid account)")
        return resp
    if resp is not None and not member.is_active:
        if FORGET_ALWAYS_TRUE:
            return rv.restStatus(request, True, msg="Password reset instructions have been sent to your email. (If valid account)")
        return resp

    if request.DATA.get("use_code", False):
        return member_forgot_password_code(request, member)

    if member.auth_code is not None and member.auth_code_expires > datetime.now():
        if FORGET_ALWAYS_TRUE:
            return rv.restStatus(request, True, msg="Password reset instructions have been sent to your email. (If valid account)")
        return rv.restPermissionDenied(request, "already sent valid auth code")

    member.auth_code = crypto.randomString(16)
    member.auth_code_expires = datetime.now() + timedelta(minutes=10)
    member.save()
    member.log("forgot", "user requested password reset", request, method="password_reset", level=17)

    token = "{}-{}".format(crypto.obfuscateID("Member", member.id), member.auth_code)
    render_to_mail("registration/password_reset_email", {
        'user': member,
        'uuid': member.uuid,
        'token': token,
        'subject': 'password reset',
        'to': [member.email],
    })

    return rv.restStatus(request, True, msg="Password reset instructions have been sent to your email. (If valid account)")


def member_forgot_password_code(request, member):
    member.generateAuthCode(6)
    code = "{} {}".format(member.auth_code[:3], member.auth_code[3:])

    context = helpers.getContext(
        request,
        user=member,
        code=code)

    if member.notify(
            context=context,
            email_only=False,
            force=True,
            subject="Login Code",
            template=settings.get("EMAIL_TEMPLATE_RESET", "email/reset_code.html"),
            sms_msg="Your login code is:\n{}".format(code)):
        member.log("requested", "user requested password reset code", request, method="login_token", level=8)
        return rv.restStatus(request, True)
    request.member = member
    member.log("error", "No valid email/phone, check users profile!", request, method="login_token", level=6)
    return rv.restPermissionDenied(request, error="No valid email/phone, check users profile!")


# time based one time passwords
@rd.urlGET('totp/qrcode')
@rd.login_required
def totp_qrcode(request):
    token = request.member.getProperty("totp_token", category="secrets", default=None)
    reset = request.DATA.get("force_reset", False)
    if token is not None and not reset:
        return rv.restPermissionDenied(request, "token exists")
    params = dict(data=request.member.totp_getURI())
    error = request.DATA.get("error", None)
    if error is not None:
        params["error"] = error
    version = request.DATA.get("version", None)
    if version is not None:
        params["version"] = int(version)
    img_format = request.DATA.get("format", "png")
    if img_format is not None:
        params["img_format"] = img_format
    scale = request.DATA.get("scale", 4)
    if scale is not None:
        params["scale"] = int(scale)
    code = generateQRCode(**params)
    if img_format == "base64":
        return HttpResponse(code, content_type="text/plain")
    elif img_format == "svg":
        return HttpResponse(code, content_type="image/svg+xml")
    return HttpResponse(code, content_type="image/png")


# time based one time passwords
@rd.urlPOST(r'^totp/verify$')
@rd.login_required
def totp_verify(request):
    code = request.DATA.get("code", None)
    if code is None or len(code) != 6:
        return rv.restPermissionDenied(request, "invalid code format")
    if not request.member.totp_verify(code):
        return rv.restPermissionDenied(request, "invalid code")
    request.member.setProperty("totp_verified", 1)
    return rv.restStatus(request, True)





# Copyright (C) 2011 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for reading OAuth 2.0 client secret files.

A client_secrets.json file contains all the information needed to interact with
an OAuth 2.0 protected service.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'


from .anyjson import simplejson

# Properties that make a client_secrets.json file valid.
TYPE_WEB = 'web'
TYPE_INSTALLED = 'installed'

VALID_CLIENT = {
    TYPE_WEB: {
        'required': [
            'client_id',
            'client_secret',
            'redirect_uris',
            'auth_uri',
            'token_uri',
        ],
        'string': [
            'client_id',
            'client_secret',
        ],
    },
    TYPE_INSTALLED: {
        'required': [
            'client_id',
            'client_secret',
            'redirect_uris',
            'auth_uri',
            'token_uri',
        ],
        'string': [
            'client_id',
            'client_secret',
        ],
    },
}


class Error(Exception):
  """Base error for this module."""
  pass


class InvalidClientSecretsError(Error):
  """Format of ClientSecrets file is invalid."""
  pass


def _validate_clientsecrets(obj):
  if obj is None or len(obj) != 1:
    raise InvalidClientSecretsError('Invalid file format.')
  client_type = list(obj.keys())[0]
  if client_type not in list(VALID_CLIENT.keys()):
    raise InvalidClientSecretsError('Unknown client type: %s.' % client_type)
  client_info = obj[client_type]
  for prop_name in VALID_CLIENT[client_type]['required']:
    if prop_name not in client_info:
      raise InvalidClientSecretsError(
        'Missing property "%s" in a client type of "%s".' % (prop_name,
                                                           client_type))
  for prop_name in VALID_CLIENT[client_type]['string']:
    if client_info[prop_name].startswith('[['):
      raise InvalidClientSecretsError(
        'Property "%s" is not configured.' % prop_name)
  return client_type, client_info


def load(fp):
  obj = simplejson.load(fp)
  return _validate_clientsecrets(obj)


def loads(s):
  obj = simplejson.loads(s)
  return _validate_clientsecrets(obj)


def _loadfile(filename):
  try:
    fp = file(filename, 'r')
    try:
      obj = simplejson.load(fp)
    finally:
      fp.close()
  except IOError:
    raise InvalidClientSecretsError('File not found: "%s"' % filename)
  return _validate_clientsecrets(obj)


def loadfile(filename, cache=None):
  """Loading of client_secrets JSON file, optionally backed by a cache.

  Typical cache storage would be App Engine memcache service,
  but you can pass in any other cache client that implements
  these methods:
    - get(key, namespace=ns)
    - set(key, value, namespace=ns)

  Usage:
    # without caching
    client_type, client_info = loadfile('secrets.json')
    # using App Engine memcache service
    from google.appengine.api import memcache
    client_type, client_info = loadfile('secrets.json', cache=memcache)

  Args:
    filename: string, Path to a client_secrets.json file on a filesystem.
    cache: An optional cache service client that implements get() and set()
      methods. If not specified, the file is always being loaded from
      a filesystem.

  Raises:
    InvalidClientSecretsError: In case of a validation error or some
      I/O failure. Can happen only on cache miss.

  Returns:
    (client_type, client_info) tuple, as _loadfile() normally would.
    JSON contents is validated only during first load. Cache hits are not
    validated.
  """
  _SECRET_NAMESPACE = 'oauth2client:secrets#ns'

  if not cache:
    return _loadfile(filename)

  obj = cache.get(filename, namespace=_SECRET_NAMESPACE)
  if obj is None:
    client_type, client_info = _loadfile(filename)
    obj = {client_type: client_info}
    cache.set(filename, obj, namespace=_SECRET_NAMESPACE)

  return next(iter(list(obj.items())))



# from django.conf.urls import *
from django.conf import settings
from django.urls import path, include, re_path
import pkgutil
import importlib
import sys
import traceback
from . import views


def loadModule(mod):
    if pkgutil.find_loader(mod) is not None:
        return importlib.import_module(mod)
    return None


urlpatterns = [
    re_path(r'^docs$', views.showDocs),
    # re_path(r'^upload$', views.chunkUploadView),
    re_path('offline', views.__offline)
]


def load_app(app, root_module=None):
    module = None
    try:
        module = loadModule(f"{app}.rpc")
        if module is None:
            module = loadModule(f"{app}.rurls")
            if module is None:
                module = loadModule(f"{app}.rapi")
    except ImportError as err:
        print("**** failed to load {0}.rpc! ****".format(app))
        print("**** missing dependencies ****")
        print("**** {0} ****".format(err))
    except SyntaxError:
        print("\t{0}: fail".format(app))
        print("Exception in user code:")
        print('-' * 60)
        traceback.print_exc(file=sys.stdout)
        print('-' * 60)
    except Exception:
        print("\t{0}: fail".format(app))
        print("Exception in user code:")
        print('-' * 60)
        traceback.print_exc(file=sys.stdout)
        print('-' * 60)
    if module:
        if not root_module:
            root_module = module
        prefix = getattr(module, 'URL_PREFIX', app.split('.')[-1])
        if len(prefix) > 1:
            prefix += "/"
        urls = path(prefix, include(root_module))
        urlpatterns.append(urls)
    return module


for app in settings.INSTALLED_APPS:
    module = load_app(app)
    if module:
        if hasattr(module, "RPC_MODULES"):
            sub_modules = getattr(module, "RPC_MODULES")
            for m in sub_modules:
                load_app(m, module)







# Generated by Django 4.1.3 on 2022-12-17 18:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('inbox', '0002_alter_message_cc'),
    ]

    operations = [
        migrations.AddField(
            model_name='attachment',
            name='content_type',
            field=models.CharField(default=None, max_length=128, null=True),
        ),
    ]



from rest import settings
from rest import helpers as rh
import socket
import importlib

IP_LOOKUP_PROVIDERS = settings.get("IP_LOOKUP_PROVIDERS", [
    "ipinfo",
    "ipapi",
    "abstractapi",
])


def getLocation(ip):
    for key in IP_LOOKUP_PROVIDERS:
        provider = importlib.import_module(f"location.providers.iplookup.{key}")
        try:
            rsp = provider.getLocation(ip)
            float(rsp.latitude)
            return rsp
        except Exception:
            rh.log_exception(f"{key}.getLocation")
    return None


def isIP(ip):
    try:
        socket.inet_aton(ip)
        return True
    except Exception:
        pass
    return False


def dnsToIP(name):
    try:
        return socket.gethostbyname(name)
    except Exception:
        pass
    return None


import threading
from concurrent import futures

# from redis import ConnectionPool, StrictRedis
from ws4redis.redis import getRedisClient, getPoolStatus
from objict import nobjict


from .models import redis, Task, TASK_STATE_STARTED
from .transports import email, http, sftp, sms, s3
from rest import helpers
from rest.log import getLogger
from rest import settings
import time


TQ_WORKERS = settings.get("TQ_WORKERS", 4)
TQ_SUBSCRIBE = settings.get("TQ_SUBSCRIBE", [])

logger = None

# use threads or processes
# because tasks are general heavy IO bound threads are probably more efficient
USE_THREADS = True

import multiprocessing
CPU_COUNT = multiprocessing.cpu_count()

if not USE_THREADS:
    # only use the number of cpus we have if using process
    TQ_WORKERS = min(TQ_WORKERS, CPU_COUNT)

class WorkManager(object):
    def __init__(self, worker_count=TQ_WORKERS, subscribe_to=TQ_SUBSCRIBE, **kwargs):
        self.worker_count = worker_count
        self.subscribe_to = subscribe_to
        self.logger = kwargs.get("logger", None)
        self.service = kwargs.get("service", None)
        self.client = None
        self.pubsub = None
        self._scheduled_tasks = {}
        self._running_count = 0
        self._pending_count = 0
        self.is_running = False
        self.host_channel = f"tq:host:{settings.HOSTNAME}"
        self.lock = threading.RLock()
        if not self.logger:
            self.logger = getLogger("root", filename="tq_worker.log")
        self.logger.info("starting manager, workers: {}".format(self.worker_count))
        self.logger.info("handling: {}".format(self.subscribe_to))
        if USE_THREADS:
            self._pool = futures.ThreadPoolExecutor(max_workers=self.worker_count)
        else:
            self._pool = futures.ProcessPoolExecutor(max_workers=self.worker_count)

    def updateCounts(self):
        self.logger.info("running: {} --- pending: {}".format(self._running_count, self._pending_count))

    def addTask(self, task):
        if task.is_stale:
            self.logger.warning("task({}) is now stale".format(task.id))
            task.failed("stale")
            return
        if task.id in self._scheduled_tasks:
            self.logger.error("task({}) is alrleady scheduled".format(task.id))
            return
        task.manager = self
        with self.lock:
            task.worker_running = False
            self._scheduled_tasks[task.id] = task
            self._pending_count += 1
            task.future = self._pool.submit(self.on_run_task, task)
        self.updateCounts()

    def addEvent(self, event):
        # self.logger.info("processing event", event)
        if event.type == "subscribe":
            # confirmation we subscribed
            self.logger.info("succesfully subscribed to: {}".format(event.channel))
            return

        self.logger.info(f"new_event@{event.channel}")

        if event.channel == "tq_restart":
            self.restart()
            return
        
        if event.data:
            event.data = nobjict.fromJSON(event.data)
        
        if event.channel == self.host_channel:
            self.on_host_event(event)
            return

        try:
            task = Task.FromEvent(event)
        except Exception:
            # this most likely means the db connection is broken!
            self.logger.exception("FromEvent")
            # recommend a restart
            self.logger.warning("db error? forcing a restart...")
            self.restart()
            return

        if not task:
            self.logger.warning("event has no task", event)
            return
        if event.channel == "tq_cancel":
            self.logger.info("cancel request received")
            try:
                self.cancelTask(task, event.data.reason)
            except Exception:
                self.logger.exception("during cancelTask")
            return
        self.addTask(task)

    def cancelTask(self, task, reason=None):
        cached_task = self._scheduled_tasks.get(task.id, None)
        if not cached_task:
            # task is not scheduled
            self.logger.warning("canceling non scheduled task({})".format(task.id))
            task.state = -2
            task.reason = reason
            task.save()
            return
        if not hasattr(cached_task, "future"):
            self.removeTask(cached_task)
            self.logger.error("task has no future!")
            return
        task = cached_task
        if task.future.running():
            # right now we don't support canceling a running task but we will try!
            self.logger.warning("attempting to stop running task({})".format(task.id))
            if self.killWorker(task._thread_id):
                time.sleep(2.0)
                if task.future.done():
                    self.logger.info("succesfully killed task({}@{})".format(task.id, task._thread_id))
                    task.state = -2
                    task.reason = reason
                    task.save()
                else:
                    self.logger.warning("failed to kill worker")
            else:
                self.logger.warning("failed to kill worker")
        else:
            if task.future.cancel():
                self.logger.info("succesfully canceled task({})".format(task.id))
                task.state = -2
                task.reason = reason
                task.save()
                self.removeTask(task)
                return
            else:
                self.logger.error("failed to cancel task({})".format(task.id))

    def killWorker(self, thread_id):
        import ctypes
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, ctypes.py_object(SystemExit))
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
            return False
        return True

    def removeTask(self, task):
        try:
            with self.lock:
                del self._scheduled_tasks[task.id]
                self._pending_count -= 1
                self.updateCounts()
        except Exception:
            pass

    def processBacklog(self):
        tasks = Task.objects.filter(state__in=[0, 1, 2])
        for task in tasks:
            if task.channel in self.subscribe_to:
                if task.cancel_requested:
                    self.logger.info("task has cancel request {}".format(task.id))
                    task.state = -2
                    if not task.reason:
                        task.reason = "task canceled"
                    task.save()
                    continue
                self.logger.debug("resubmitting job {}".format(task.id))
                self.addTask(task)
            else:
                self.logger.warning("ignore job {}:{}".format(task.id, task.channel))

    def _on_webrequest(self, task):
        if http.REQUEST(task):
            task.completed()
        elif task.attempts < task.max_attempts:
            # -1 will auto calculate retry with back off
            # lets report the issue
            task.notifyError(reason=task.reason)
            task.retry_later(from_now_secs=-1)
        else:
            task.failed("max attempts")

    def _on_hookrequest(self, task):
        if task.model == "tq_email_request":
            resp = email.SEND(task)
        elif task.model == "tq_sms_request":
            resp = sms.SEND(task)
        elif task.model == "tq_sftp_request":
            resp = sftp.SEND(task)
        elif task.model == "tq_s3_request":
            resp = s3.UPLOAD(task)
        if resp:
            task.completed()
        elif task.attempts < task.max_attempts:
            task.retry_later(from_now_secs=-1)
        else:
            task.failed()

    def on_host_event(self, event):
        action = event.data.action
        if action == "ping":
            redis.publish(
                event.data.response_channel,
                dict(action="pong", hostname=settings.HOSTNAME))
        elif action == "restart":
            redis.publish(
                event.data.response_channel,
                dict(action="restarting", hostname=settings.HOSTNAME))
            self.restart() 
        elif action == "get_stats":
            data = nobjict(action="stats", hostname=settings.HOSTNAME)
            data.uptime = time.time() - self.started_at
            data.subscribed = self.subscribe_to
            data.running = self._running_count
            data.pending = self._pending_count
            data.scheduled = len(self._scheduled_tasks.keys())
            data.pool = dict(**getPoolStatus())
            redis.publish(event.data.response_channel, data)

    def on_task_started(self, task):
        with self.lock:
            task.worker_running = True
            self._running_count += 1
            self._pending_count -= 1
            self.updateCounts()

    def on_task_ended(self, task):
        with self.lock:
            task.worker_running = False
            del self._scheduled_tasks[task.id]
            self._running_count -= 1
            self.updateCounts()

    def on_run_task(self, task):
        self.logger.info("running task({})".format(task.id))

        # check if its canceled
        try:
            self.on_task_started(task)
            task.refresh_from_db()
            task._thread_id = threading.current_thread().ident
            self.logger.debug("running on thread:{}".format(task._thread_id))
            if task.state not in [0, 1, 2, 10] or task.cancel_requested:
                self.logger.info("task({}) was canceled?".format(task.id))
                self.on_task_ended(task)
                return

            if task.is_stale:
                self.logger.warning("task({}) is now stale".format(task.id))
                task.failed("stale")
                self.on_task_ended(task)
                return
        except Exception as err:
            self.logger.exception(err)
            return

        handler = None
        if task.model == "tq_web_request":
            handler = self._on_webrequest
        elif task.model in ["tq_sftp_request", "tq_email_request", "tq_sms_request", "tq_s3_request"]:
            handler = self._on_hookrequest
        else:
            try:
                handler = task.getHandler()
            except Exception as err:
                self.logger.exception(err)
                task.log_exception(err)
                task.failed(str(err))
                self.on_task_ended(task)
                return

        if handler is None:
            self.logger.error("failed to find handler: task({})".format(task.id))
            task.failed("failed to find handler")
            self.on_task_ended(task)
            return

        # self.logger.debug("task.started()")
        task.started()
        try:
            # self.logger.debug("task({}) calling handler".format(task.id))
            handler(task)
            if task.state == TASK_STATE_STARTED:
                task.completed()
            # self.logger.debug("task({}) handler finished".format(task.id))
        except Exception as err:
            self.logger.exception(err, "task({}) had exception".format(task.id))
            task.log_exception(err)
            if "connection already closed" in str(err).lower():
                # this is a nasty little bug in django when forking django db connections
                # we will schedule the task to retry later
                task.retry_later()
                # let us try and close db connections?
                hack_closeDjangoDB()
                # or should we restart the task queue?
            else:
                task.failed(str(err))
        except SystemExit:
            self.logger.error("task({}) was killed".format(task.id))
        finally:
            self.on_task_ended(task)
        self.logger.info("task({}) finished with state {}".format(task.id, task.state))

    def run_forever(self):
        self.logger.info("starting work manager...")
        self.__open()
        self.logger.info("listening for incoming events...")
        while self.is_running:
            for event in self.pubsub.listen():
                if self.is_running:
                    # self.logger.debug("new event", event)
                    try:
                        event = nobjict.fromdict(event)
                        event.channel = helpers.toString(event.channel)
                        self.addEvent(event)
                    except Exception as err:
                        self.logger.exception(err)
        self.__close()

    def restart(self):
        if self.service:
            self.is_running = False
            self.stop(timeout=30.0)
            self.service.restart()

    def __open(self):
        if self.client is not None:
            return
        self.started_at = time.time()
        self.is_running = True
        self.client = getRedisClient()
        self.pubsub = self.client.pubsub()
        self.__subscribe()

    def __subscribe(self):
        if self.host_channel not in self.subscribe_to:
            self.subscribe_to.append(self.host_channel)
        for key in self.subscribe_to:
            self.logger.info("subscribing to: {}".format(key))
            self.pubsub.subscribe(key)
        self.pubsub.subscribe("tq_cancel")
        self.pubsub.subscribe("tq_restart")
        self.client.sadd("tq:host:online", settings.HOSTNAME)

    def __unsubscribe(self):
        self.client.srem("tq:host:online", settings.HOSTNAME)
        for key in self.subscribe_to:
            self.pubsub.unsubscribe(key)
        self.pubsub.unsubscribe("tq_cancel")
        self.pubsub.unsubscribe("tq_restart")

    def __close(self):
        if self.client is None:
            return

        self.__unsubscribe()
        self.client = None
        self.pubsub = None
        self.logger.info("closed")

    def stop(self, timeout=30.0):
        self.updateCounts()
        self.logger.info("stopping, canceling pending tasks...")
        self.is_running = False
        # we need to cancel all futures not running:
        try:
            redis.publish("tq_cancel", {"pk":1})
            with self.lock:
                self.updateCounts()
                for key, task in list(self._scheduled_tasks.items()):
                    if not hasattr(task, "future"):
                        continue
                    if not task.future.running():
                        task.future.cancel()
                self.updateCounts()
        except Exception as err:
            self.logger.exception(err)

        self.logger.info("waiting for {} running tasks, timeout: {}".format(self._running_count, timeout))
        time.sleep(1.0)
        self.__close()
        timeout_at = time.time() + timeout
        while self._running_count > 0 and time.time() < timeout_at:
            # we are waiting for all jobs to finish
            time.sleep(1.0)

        self.updateCounts()
        if self._running_count:
            self.logger.error("timedout waiting for long running tasks, stopping anyway")
            # lets set all to failed that are running
            for pk, task in list(self._scheduled_tasks.items()):
                if task.worker_running:
                    if task.current_runtime >= 300:
                        task.failed("killed because task is taking to long to run")
                    else:
                        task.retry_later("worker engine restarting")
        else:
            self.logger.info("all tasks complete and workers stopped!")


def hack_closeDjangoDB():
    from django.db import connections
    for conn in connections.all():
        conn.close_if_unusable_or_obsolete()





# Generated by Django 4.2.11 on 2024-08-13 04:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wiki', '0002_alter_pagemedia_entry'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='perms',
            field=models.CharField(db_index=True, default=None, max_length=255, null=True),
        ),
    ]



from rest import decorators as rd
from rest import views as rv
from rest import helpers as rh
from rest import settings
from . import models as am
from .extra import abuse 
from .parsers import ossec
from taskqueue.models import Task
import incident

LOG_REST_PREFIX = settings.get("REST_PREFIX", "api/")
if not LOG_REST_PREFIX.startswith("/"):
    LOG_REST_PREFIX = f"/{LOG_REST_PREFIX}"
STATUS_ON_PERM_DENIED = settings.get("STATUS_ON_PERM_DENIED", 403)


def patched_restPermissionDenied(request, error="permission denied", 
                                 error_code=403, status=STATUS_ON_PERM_DENIED,
                                 component=None, component_id=None):
    
    description = f"permission denied: {error_code} '{error}' for {request.user} {request.method}:{request.path}"
    # rh.log_error(description)
    if error_code == 404:
        if not request.path.startswith(LOG_REST_PREFIX) and not request.path.startswith("/rpc"):
            # just ignore these
            return rv.restResult(request, dict(status=False, error=error, error_code=error_code), status=status)

    metadata = dict(error=error, error_code=error_code, details=description)
    if component is not None:
        metadata["component"] = component
        metadata["component_id"] = component_id
    if hasattr(request, "member") and request.member is not None:
        request.DATA.log()
        request.member.log("rest_denied", error, request, level=10)
        if component is None:
            metadata["component"] = "account.Member"
            metadata["component_id"] = request.member.pk

    if getattr(request, "token_bearer", None) is not None:
        metadata["token_bearer"] = request.token_bearer
        metadata["token"] = f"***{request.token[-6:]}"
    incident.event_now(
        "rest_denied",
        description,
        level=10,
        request=request,
        **metadata)
    return rv.restResult(request, dict(status=False, error=error, error_code=error_code), status=status)


def patched_restNotFound(request):
    return rv.restPermissionDenied(request, error="not found", error_code=404)


if settings.REPORT_PERMISSION_DENIED:
    rv.restPermissionDenied = patched_restPermissionDenied


@rd.urlPOST(r'^ossec/alert$')
def ossec_alert_creat_from_request(request):
    payload = request.DATA.get("payload")
    if payload:
        try:
            # TODO make this a task (background it)
            # rh.log_error("parsing payload", payload)
            od = ossec.parseAlert(request, payload)
            # lets now create a local event
            if od is not None:
                level = 10
                if od.level > 10:
                    level = 1
                elif od.level > 7:
                    level = 2
                elif od.level == 6:
                    level = 3
                elif od.level == 5:
                    level = 4
                elif od.level == 4:
                    level = 6
                elif od.level <= 3:
                    level = 8
                metadata = od.toDict(graph="default")
                metadata.update(od.metadata)
                # we reuse the ssh_sig because it is a text field to store urls
                # ssh_sig = metadata.get("ssh_sig", None)
                # if ssh_sig is not None and ssh_sig.startswith("http"):    
                #     metadata["url"] = ssh_sig
                #     metadata["domain"] = ossec.extractDomain(ssh_sig)
                #     metadata["path"] = ossec.extractUrlPath(ssh_sig)
                #     metadata.pop("ssh_sig")
                if od.geoip:
                    metadata["country"] = od.geoip.country
                    metadata["city"] = od.geoip.city
                    metadata["province"] = od.geoip.state
                    metadata["isp"] = od.geoip.isp

                am.Event.createFromDict(None, {
                    "hostname": od.hostname,
                    "description": od.title,
                    "details": od.text,
                    "level": level,
                    "category": "ossec",
                    "component": "incident.ServerOssecAlert",
                    "component_id": od.id,
                    "reporter_ip": od.src_ip,
                    "metadata": metadata
                })
                return rv.restStatus(request, True)
        except Exception as err:
            rh.log_exception()
            stack = rh.getStackString()
            # rh.log_exception("during ossec alert", payload)
            metadata = dict(ip=request.ip, payload=payload)
            am.Event.createFromDict(None, {
                "hostname": request.get_host(),
                "description": f"error parseing alert: {err}",
                "details": stack,
                "level": 8,
                "category": "ossec_error",
                "metadata": metadata
            })
    # rh.log_error("ossec alert", request.DATA.asDict())
    return rv.restStatus(request, False, error="no alert data")


@rd.urlGET(r'^ossec$')
@rd.urlGET(r'^ossec/(?P<pk>\d+)$')
@rd.login_required
def on_ossec(request, pk=None):
    return am.ServerOssecAlert.on_rest_request(request, pk)


@rd.urlPOST(r'^event$')
def rest_on_create_event(request, pk=None):
    # TODO check for key?
    resp = am.Event.on_rest_request(request)
    return rv.restStatus(request, True)


@rd.urlGET(r'^event$')
@rd.urlGET(r'^event/(?P<pk>\d+)$')
@rd.urlPOST(r'^event/(?P<pk>\d+)$')
@rd.login_required
def rest_on_event(request, pk=None):
    return am.Event.on_rest_request(request, pk)


@rd.url(r'^incident$')
@rd.url(r'^incident/(?P<pk>\d+)$')
@rd.login_required
def rest_on_incident(request, pk=None):
    return am.Incident.on_rest_request(request, pk)


@rd.url(r'^incident/history$')
@rd.url(r'^incident/history/(?P<pk>\d+)$')
@rd.login_required
def rest_on_incident_history(request, pk=None):
    return am.IncidentHistory.on_rest_request(request, pk)


@rd.url(r'^rule$')
@rd.url(r'^rule/(?P<pk>\d+)$')
@rd.login_required
def rest_on_rule(request, pk=None):
    return am.Rule.on_rest_request(request, pk)


@rd.url(r'^rule/check$')
@rd.url(r'^rule/check/(?P<pk>\d+)$')
@rd.login_required
def rest_on_rule_check(request, pk=None):
    return am.RuleCheck.on_rest_request(request, pk)


# BEGIN FIREWALL
@rd.urlPOST(r'^ossec/firewall$')
def ossec_firewall_event(request):
    data = request.DATA.toObject()
    # rh.debug("firewall event", data)
    gip = ossec.GeoIP.lookup(data.ip)
    if gip:
        data.country = gip.country
        data.province = gip.state
        data.city = gip.city
        data.isp = gip.isp
        data.ip_hostname = gip.hostname
    if data.action == "add":
        data.action = "blocked"
    else:
        data.action = "unblocked"
    record_block(data.ip, data.action, data.hostname, data)
    # now block the ip globally
    if settings.FIREWALL_GLOBAL_BLOCK:
        Task.Publish("incident", "firewall_block", dict(ip=data.ip), channel="tq_broadcast")
    return rv.restStatus(request, True)


def record_block(ip, action, hostname, metadata):
    title = f"FIREWALL: {ip} {action} on {hostname}"
    am.Event.createFromDict(None, dict(
        hostname=hostname,
        reporter_ip=ip,
        description=title,
        details=title,
        category="firewall",
        level=6,
        metadata=metadata
    ))


@rd.urlGET(r'^firewall$')
# @rd.superuser_required
def rest_firewall_blocked(request, pk=None):
    return rv.restReturn(request, dict(data=dict(blocked=rh.getBlockedHosts())))


@rd.urlPOST(r'^firewall$')
@rd.superuser_required
@rd.requires_params(["ip", "action"])
def rest_firewall_block(request):
    # block the ip globally
    ip = request.DATA.get("ip")
    action = request.DATA.get("action")
    if action not in ["block", "unblock"]:
        return rv.restPermissionDenied(request)
    metadata = request.DATA.asDict()
    metadata.username = request.member.username
    record_block(ip, action, "all", metadata)
    Task.Publish("incident", f"firewall_{action}", dict(ip=ip), channel="tq_firewall")
    return rv.restStatus(request, True)


@rd.url('ticket')
@rd.url('ticket/<int:pk>')
def rest_on_ticket(request, pk=None):
    return am.Ticket.on_rest_request(request, pk)


@rd.urlGET('abuse/lookup')
@rd.requires_params(["ip"])
def rest_on_abuse_ip_lookup(request):
    resp = abuse.lookup(request.DATA.get("ip"), request.DATA.get("max_age", 0))
    return rv.restReturn(request, resp)


@rd.urlPOST('abuse/report')
@rd.requires_params(["ip", "comment", "categories"])
def rest_on_report_abuse(request):
    resp = abuse.reportAbuse(
        ip=request.DATA.get("ip"),
        comment=request.DATA.get("comment"),
        categories=request.DATA.get("categories"))
    return rv.restReturn(request, resp)



from django import http
from django.conf import settings
from urllib.parse import urlparse
from django.utils.cache import patch_vary_headers

if hasattr(settings, "CORS_SHARING_ALLOWED_ORIGINS"):
    CORS_SHARING_ALLOWED_ORIGINS = settings.CORS_SHARING_ALLOWED_ORIGINS
else:
    CORS_SHARING_ALLOWED_ORIGINS = '*'

if hasattr(settings, "CORS_SHARING_ALLOWED_METHODS"):
    CORS_SHARING_ALLOWED_METHODS = settings.CORS_SHARING_ALLOWED_METHODS
else:
    CORS_SHARING_ALLOWED_METHODS = ['POST', 'GET', 'OPTIONS', 'PUT', 'DELETE']

if hasattr(settings, "CORS_SHARING_ALLOWED_HEADERS"):
    CORS_SHARING_ALLOWED_HEADERS = settings.CORS_SHARING_ALLOWED_HEADERS
else:
    CORS_SHARING_ALLOWED_HEADERS = [
        'accept',
        'accept-encoding',
        'authorization',
        'content-type',
        'dnt',
        'origin',
        'user-agent',
        'x-authtoken',
        'x-csrftoken',
        'x-sessionid',
        'x-requested-with']

CORS_ALLOW_CREDENTIALS = getattr(settings, "CORS_ALLOW_CREDENTIALS", True)


class CorsMiddleware(object):
    """
        This middleware allows cross-domain XHR using the html5 postMessage API.

        Access-Control-Allow-Origin: http://foo.example
        Access-Control-Allow-Methods: POST, GET, OPTIONS, PUT, DELETE
    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # rest_helpers.log_print(dict(request.META))
        # check if preflight header, then return HTTP 200
        if request.method.lower() == "options" and 'HTTP_ACCESS_CONTROL_REQUEST_METHOD' in request.META:
            response = http.HttpResponse()
        else:
            response = self.get_response(request)
        response = self.updateResponse(request, response)
        # rest_helpers.log_print(response._headers)
        return response

    def getAllowedOrigin(self, request):
        origin = request.META.get('HTTP_ORIGIN', None)
        if origin:
            if CORS_SHARING_ALLOWED_ORIGINS == '*':
                return origin
            host = urlparse(origin)
            if 'localhost' in host:
                host = 'localhost'
            if type(CORS_SHARING_ALLOWED_ORIGINS) is list:
                if host in CORS_SHARING_ALLOWED_ORIGINS:
                    return origin
            elif type(CORS_SHARING_ALLOWED_ORIGINS) in [str, str]:
                if host == CORS_SHARING_ALLOWED_ORIGINS:
                    return origin

        host = request.META.get('HTTP_HOST', None)
        if not host:
            return None
        return "https://{}".format(host)

    def updateResponse(self, request, response):
        # rest_helpers.log_print("CORS: updating response")
        if CORS_ALLOW_CREDENTIALS:
            response['Access-Control-Allow-Credentials'] = 'true'
        else:
            response['Access-Control-Allow-Credentials'] = 'false'

        allowed_origin = self.getAllowedOrigin(request)
        if allowed_origin:
            # rest_helpers.log_print('Access-Control-Allow-Origin: {}'.format(allowed_origin))
            response['Access-Control-Allow-Origin']  = allowed_origin
            patch_vary_headers(response, ['Origin'])

        if request.method.lower() == "options":
            response['Access-Control-Allow-Headers'] = ",".join( CORS_SHARING_ALLOWED_HEADERS )
            response['Access-Control-Allow-Methods'] = ",".join( CORS_SHARING_ALLOWED_METHODS )
            response["Access-Control-Max-Age"] = "7200"
        return response




from __future__ import unicode_literals
from django.core.management.base import BaseCommand, CommandError
from redis_metrics.utils import get_r


class Command(BaseCommand):
    args = '<metric-key>'
    help = "Removes a metric and its data"

    def handle(self, *args, **options):
        if len(args) == 0:
            raise CommandError("You must provide a metric name")
        metric_slug = args[0]

        r = get_r()
        r.delete_metric(metric_slug)



from .ffmpeg import render as vid_render
from ..engines import rtmpsink
from ..exceptions import *

import tempfile
import os

from django.conf import settings

PresetConfig = {
    'name':		'Akamai streaming',
    'description':	'Stream video to akamai',
    'stage':	192,
    'use':		'rtmp',
    'applicable_kinds': 'L',
    'parameters': {
        'endpoint_id': {
            'description':	'Endpoint ID',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
        },
        'encoder_id': {
            'description':	'Encoder ID string',
            'required':	False,
            'kind':		'S',
            'configurable':	True,
        },
        'account_id': {
            'description':	'Account ID',
            'required':	True,
            'kind':		'S',
            'configurable':	True,
        },
        'account_password': {
            'description':	'Account Password',
            'required':	True,
            'kind':		'S',
            'configurable':	True,
        },
        'stream_name': {
            'description':	'Output stream name',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
        },
        'audio_bitrate': {
            'description':	'Audio bitrate (kbps)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	128,
        },
        'audio_channels': {
            'description':	'Audio channels',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	2,
        },
        'audio_freq': {
            'description':	'Audio frequency (hz)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	44100,
        },
        'video_bitrate': {
            'description':	'Video bitrate (kbps)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	512,
        },
        'frame_rate': {
            'description':	'Frame rate',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	30,
        },
        'key_frame': {
            'description':	'Key frame rate',
            'required':	False,
            'kind':		'I',
            'configurable':	True,
            'value':	25,
        },
    },
}

def cleanup(data, haserr):
    if haserr:
        ret = data['proc'].wait(nohang=True)
        if ret == None:
            data['proc'].kill()
            data['proc'].wait()
        elif ret == 0:
            ret = -1
    else:
        ret = data['proc'].wait()
    try:
        os.unlink(data['fname'])
    except:
        pass
    if ret:
        raise CmdError(ret, data['proc'].err(), cmd='*rtmpsink')
    
def render(item, renditiondef, params):
    params['video_container'] = 'flv'
    params['video_codec'] = 'libx264'
    params['audio_codec'] = 'libfaac'
    if params.get('audio_bitrate') >= 96 and 'audio_channels' not in params:
        params['audio_channels'] = 2
    (fp, fname) = tempfile.mkstemp(prefix="t_akamai_")
    os.close(fp)
    os.unlink(fname)
    os.mkfifo(fname)

    proc = rtmpsink.run(
        'rtmp://%s%d%s conn=S:encoder:%s:%s live=true akUserId=%s akPass=%s playpath=%s@%d' %  (
            getattr(settings, 'AKAMAI_ENTRY_PREFIX', 'p.ep'),
            params['endpoint_id'],
            getattr(settings, 'AKAMAI_ENTRY_SUFFIX', '.i.akamaientrypoint.net/EntryPoint'),
            params.get('encoder_id', '0.0.0.0'),
            params['account_id'],
            params['account_id'],
            params['account_password'],
            params['stream_name'],
            params['endpoint_id'],
        ), fname)
    params['stream_output'] = fname

    params['url'] = 'rtmp://%s%s%s%s@%d' % (
        getattr(settings, 'AKAMAI_EDGE_PREFIX', 'cp'),
        params['account_id'],
        getattr(settings, 'AKAMAI_EDGE_SUFFIX', '.live.edgefcs.net/live/'),
        params['stream_name'],
        params['endpoint_id'],
    )

    return vid_render(item, renditiondef, params, cleanup=cleanup, data={'proc': proc, 'fname': fname})



from django.db import models
from rest import models as rm
from location import geolocate


class GeoIPLocation(models.Model, rm.RestModel):
    # assocate with a user if possible
    owner = models.ForeignKey("account.User", blank=True, null=True, default=None, related_name="locations", on_delete=models.CASCADE)

    # use this to reference a Model
    obj_id = models.IntegerField(blank=True, default=0)
    obj_kind = models.CharField(max_length=64, blank=True, null=True, default="")

    ip = models.CharField(max_length=32)
    created = models.DateTimeField(auto_now_add=True)
    city = models.CharField(max_length=64, blank=True, null=True)
    region = models.CharField(max_length=64)
    country = models.CharField(max_length=64)
    area_codes = models.CharField(max_length=64)
    timezone = models.CharField(max_length=32)
    kind = models.CharField(max_length=32)

    lat = models.FloatField(default=0.0, blank=True)
    lng = models.FloatField(default=0.0, blank=True)

    @staticmethod
    def cloneForOwner(obj, owner):
        clone = obj
        clone.pk = None
        clone.id = None
        clone.owner = owner
        clone.save()
        return clone

    @staticmethod
    def createByIP(ip, owner=None):
        # first check if we have already seen this ip
        locs = GeoIPLocation.objects.filter(ip=ip)
        if locs.count():
            if locs.filter(owner=owner).count():
                return locs.filter(owner=owner).first()
            # clone the first obj
            return GeoIPLocation.cloneForOwner(locs.first(), owner)

        res = geolocate.locateIP(ip)
        obj = None
        if res:
            obj = GeoIPLocation(owner=owner, ip=ip, lat=res["lat"], lng=res["lng"])
            if "city" in res:
                obj.city = res["city"]
            if "country" in res:
                obj.country = res["country"]
            if "region" in res:
                obj.region = res["region"]
            obj.save()
        return obj



# Generated by Django 3.1.5 on 2021-01-05 21:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('sessionlog', '0001_initial'),
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PersistentLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('when', models.DateTimeField(auto_now_add=True, db_index=True, help_text='When change was logged')),
                ('level', models.IntegerField(default=0)),
                ('message', models.TextField(blank=True, default=None, null=True)),
                ('action', models.TextField(blank=True, db_index=True, default=None, max_length=127, null=True)),
                ('component', models.CharField(blank=True, db_index=True, default=None, max_length=127, null=True)),
                ('pkey', models.IntegerField(blank=True, db_index=True, default=None, null=True)),
                ('request_method', models.CharField(blank=True, default=None, max_length=254, null=True)),
                ('tid', models.CharField(blank=True, db_index=True, default=None, max_length=254, null=True)),
                ('request_path', models.TextField(blank=True, default=None, null=True)),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='account.group')),
                ('session', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='sessionlog.sessionlog')),
                ('user', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='AuditLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('when', models.DateTimeField(auto_now_add=True, help_text='When change was logged')),
                ('model', models.CharField(db_index=True, max_length=127)),
                ('pkey', models.IntegerField(db_index=True)),
                ('attribute', models.CharField(max_length=127)),
                ('how', models.TextField(blank=True, max_length=127, null=True)),
                ('referer', models.TextField(blank=True, null=True)),
                ('stack', models.TextField(blank=True, null=True)),
                ('oldval', models.TextField(blank=True, null=True)),
                ('newval', models.TextField(blank=True, null=True)),
                ('session', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='sessionlog.sessionlog')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('can_read', 'Can read audit log'),),
            },
        ),
    ]



from objict import objict

META_KEYS = ["SERVER_PROTOCOL", "REQUEST_METHOD", "QUERY_STRING", "HTTP_USER_AGENT"]


def _request_to_meta(request, metadata):
    metadata.ip = request.ip
    if "path" not in metadata:
        metadata.path = request.path
    for key in META_KEYS:
        value = request.META.get(key, None)
        if value is not None and isinstance(value, str) and "." not in key:
            metadata[key.lower()] = str(value)
    if hasattr(request, "buid"):
        metadata["buid"] = request.buid
    if "username" not in metadata and hasattr(request, "member") and request.member is not None:
        metadata["username"] = request.member.username
    return metadata


def event(category, description, level=10, request=None, group=None, **kwargs):
    from taskqueue.models import Task
    data = objict(category=category, description=description, level=level, group=group)
    data.metadata = objict.fromdict(kwargs)
    if request is not None:
        _request_to_meta(request, data.metadata)
    Task.Publish("incident", "new_event", channel="tq_app_handler", data=data)


def event_now(category, description, level=10, request=None, group=None, **kwargs):
    from .models.event import Event
    data = objict(category=category, description=description, level=level, group=group)
    data.metadata = objict.fromdict(kwargs)
    if request is not None:
        _request_to_meta(request, data.metadata)
    if "hostname" in data.metadata:
        data.hostname = data.metadata.hostname
    if "details" in data.metadata:
        data.details = data.metadata.details
    if "component" in data.metadata:
        data.component = data.metadata.component
    if "component_id" in data.metadata:
        data.component_id = data.metadata.component_id
    if "ip" in data.metadata:
        data.reporter_ip = data.metadata.ip
    Event.createFromDict(None, data)


def exception_event(request, error, body=None, stack=None, 
                    title="REST Error", level=1,
                    category="rest_exception", **kwargs):
    from rest import settings
    import traceback
    if stack is None:
        stack = str(traceback.format_exc())
    host = request.get_host()
    server = settings.get("HOSTNAME", "unknown")
    if body is None:
        try:
            body = request.body.decode('utf-8')
        except Exception:
            body = request.DATA.asDict()
    description = kwargs.pop("description", error)
    subject = f"{title} on {host} : {error}"
    username = "anonymous"
    if request.member:
        username = request.member.username

    metadata = {
        "method": request.method,
        "path": request.path,
        "server": server,
        "error": str(error),
        "username": username,
        "ip": request.ip,
    }

    for k, v in kwargs.items():
        metadata[k] = v

    if request.auth_model is not None:
        metadata['auth_model'] = str(request.auth_model)
    if request.group is not None:
        metadata['group_name'] = request.group.name
        metadata['group_id'] = request.group.id
    if hasattr(request, "buid"):
        metadata["buid"] = request.buid
    if hasattr(request, "_log_component"):
        metadata["component"] = request._log_component
        metadata["pkey"] = request._log_pk

    event_now(
        category, description=subject, level=level, hostname=host,
        details=stack, reporter_ip=request.ip, request=request,
        **metadata)



# Generated by Django 4.0.2 on 2022-02-22 01:26

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PhonenumberInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('number', models.CharField(db_index=True, max_length=200)),
                ('owner_name', models.CharField(default=None, max_length=200, null=True)),
                ('owner_kind', models.CharField(default=None, max_length=200, null=True)),
                ('country', models.CharField(default=None, max_length=64, null=True)),
                ('carrier_name', models.CharField(default=None, max_length=64, null=True)),
                ('kind', models.CharField(default=None, max_length=64, null=True)),
                ('is_valid', models.BooleanField(default=False)),
                ('is_verified', models.BooleanField(default=False)),
                ('sms_allowed', models.BooleanField(default=False)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='SMS',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sid', models.CharField(blank=True, default=None, max_length=125, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('transport', models.CharField(blank=True, default=None, max_length=125, null=True)),
                ('endpoint', models.CharField(blank=True, db_index=True, default=None, max_length=200, null=True)),
                ('srcpoint', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('is_inbound', models.BooleanField(default=False)),
                ('status', models.CharField(db_index=True, default=None, max_length=64, null=True)),
                ('reason', models.TextField(default=None, null=True)),
                ('message', models.TextField()),
                ('by', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='telephony_sms_out', to='account.member')),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='telephony_sms', to='account.group')),
                ('to', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='telephony_sms_in', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



from django.db import models

from rest import models as rm
from rest.extra import JSONMetaData
from rest import log
from rest import settings
from objict import objict

import metrics

from datetime import datetime, timedelta
from .incident import Incident, INCIDENT_STATE_PENDING
from .rules import Rule

INCIDENT_METRICS = settings.get("INCIDENT_METRICS", False)
INCIDENT_EVENT_METRICS = settings.get("INCIDENT_EVENT_METRICS", False)
INCIDENT_EVENT_GRANULARITY = settings.get("INCIDENT_EVENT_GRANULARITY", "hourly")
EVENT_TO_INCIDENT_LEVEL = settings.get("EVENT_TO_INCIDENT_LEVEL", 4)
EVENT_DETAIL_TEMPLATES = settings.get("EVENT_DETAIL_TEMPLATES", None)
EVENT_META_KEYWORDS = settings.get("EVENT_META_KEYWORDS", [
        "path", "ip", "reporter_ip", "code", 
        "reason", "buid", "merchant", "tid", 
        "group", "http_user_agent", "user_agent",
        "app_url", "isp", "city", "state", "country",
        "username"
    ])

logger = log.getLogger("incident", filename="incident.log")

"""
very generic 
external system can post an event
{
     "description": "Critical Test Event",
     "hostname": "r1",
     "details": "A critical event occurred on r1 running blah blah",
     "level": 7,
     "category": "prepaid.event",
     "metadata": {
        "error_stack": "....."
     }
}
"""


class Event(JSONMetaData, rm.RestModel):
    class RestMeta:
        POST_SAVE_FIELDS = ["level", "catagory"]
        SEARCH_FIELDS = ["description", "hostname"]
        VIEW_PERMS = ["view_incidents", "view_logs"]
        CREATE_PERMS = None  # allow anyone to create an event
        GRAPHS = {
            "default": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic"
                },
            },
            "detailed": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic",
                    "generic__component": "basic",
                },
            },
        }

    created = models.DateTimeField(auto_now_add=True)
    reporter_ip = models.CharField(max_length=16, blank=True, null=True, default=None, db_index=True)

    hostname = models.CharField(max_length=255, blank=True, null=True, default=None, db_index=True)
    description = models.CharField(max_length=84)
    details = models.TextField(default=None, null=True)

    level = models.IntegerField(default=0, db_index=True)
    category = models.CharField(max_length=124, db_index=True)
    # code = models.IntegerField(default=0, db_index=True)

    group = models.ForeignKey(
        "account.Group", on_delete=models.SET_NULL, 
        related_name="+", null=True, default=None)
 
    component = models.SlugField(max_length=250, null=True, blank=True, default=None)
    component_id = models.IntegerField(null=True, blank=True, default=None)

    # this allows us to bundle multiple events to an incident
    incident = models.ForeignKey(
        Incident, null=True, default=None, 
        related_name="events", on_delete=models.SET_NULL)

    def runRules(self):
        for rule in Rule.objects.filter(category=self.category).order_by("priority"):
            if rule.run(self):
                return rule
        return None

    @property
    def details_by_category(self):
        # returns detailed text based on the category settings
        # if EVENT_DETAIL_TEMPLATES is None or self.category not in EVENT_DETAIL_TEMPLATES:
        #     return self.details
        output = []
        if self.component:
            output.append(f"{self.component}({self.component_id})")
        for key in EVENT_META_KEYWORDS:
            if self.metadata.get(key, None) is not None:
                output.append(f"{key}: {self.metadata[key]}")
        if self.details:
            output.append(self.details)
            output.append("")
            output.append("")
        return "\n".join(output)

    def lookupIP(self, ip):
        GeoIP = rm.RestModel.getModel("location", "GeoIP")
        gip = GeoIP.lookup(ip)
        self.setProperty("ip", ip)
        if gip:
            self.setProperty("country", gip.country)
            self.setProperty("state", gip.state)
            self.setProperty("city", gip.city)
            self.setProperty("isp", gip.isp)

    def set_description(self, value):
        # trun desc to 84
        if value is None:
            value = ""
        if len(value) > 84:
            value = value[:80] + "..."
        self.description = value

    def on_rest_pre_save(self, request):
        if self.hostname is None:
            self.hostname = settings.HOSTNAME
        if not self.getProperty("hostname", None):
            self.setProperty("hostname", self.hostname)
        self.setProperty("category", self.category)
        if request and request.DATA.get("ip_lookup", field_type=bool):
            self.reporter_ip = request.ip
            self.lookupIP(request.ip)

    def on_rest_saved(self, request, is_new=False):
        if not is_new:
            return
        if INCIDENT_EVENT_METRICS:
            if self.hostname:
                metrics.metric(f"incident_evt_{self.hostname}", category="incident_events", min_granularity="hourly")
            metrics.metric("incident_evt", min_granularity=INCIDENT_EVENT_GRANULARITY)

        self.setProperty("level", self.level)
        if request is not None:
            self.reporter_ip = request.ip
        # run through rules for the category
        hit_rule = self.runRules()
        priority = 10
        incident = None
        action_count = 0
        if hit_rule is not None:
            # logger.error(f"RULE HIT: {hit_rule.name}")
            priority = hit_rule.priority
            if hit_rule.action == "ignore":
                self.save()
                return
            if hit_rule.bundle > 0:
                incident = Incident.getBundled(rule=hit_rule, event=self)

        elif self.level >= EVENT_TO_INCIDENT_LEVEL:
            # we ignore levels 4 and higher if they did not create a rule
            self.save()
            # logger.info(f"ignore event {self.pk} {self.description}")
            return

        # always create an incident 
        if incident is None:
            incident = Incident(
                rule=hit_rule, priority=priority,
                reporter_ip=self.reporter_ip,
                category=self.category,
                group=self.group,
                component=self.component, 
                component_id=self.component_id,
                hostname=self.hostname)
            if self.group is None and hit_rule is not None:
                incident.group = hit_rule.group
            if hit_rule is not None and hit_rule.action_after != 0:
                incident.state = INCIDENT_STATE_PENDING
            # TODO possibly make this smarter?
            if hit_rule and hit_rule.title_template and "{" in hit_rule.title_template:
                try:
                    incident.description = hit_rule.title_template.format(event=self)
                except Exception:
                    logger.exception(hit_rule.title_template)
                    incident.description = self.description
            elif self.category == "ossec":
                incident.description = f"{self.hostname}: {self.description}"
                incident.save()
                incident.updateAbuseInfo()
            else:
                incident.description = self.description
            incident.save()
            incident.updateMeta(self)
        self.incident = incident
        self.save()
        if INCIDENT_METRICS:
            # we want to track any incidents, including bundles
            if self.hostname:
                metrics.metric(f"incidents_{self.hostname}", category="incidents", min_granularity="hourly")
            metrics.metric("incidents", min_granularity=INCIDENT_EVENT_GRANULARITY)

        try:
            incident.triggerAction()
        except Exception:
            logger.exception()




from django.contrib import admin
from . import models as location
from django import forms


admin.site.register(location.GeoLocation)
admin.site.register(location.GeoPosition)
admin.site.register(location.GeoIP)
# admin.site.register(location.GeoIPLocation)
admin.site.register(location.Address)





from rest.decorators import periodic, PERIODIC_EVERY_5_MINUTES
from rest import helpers
from rest.log import getLogger
from taskqueue.models import Task, TASK_STATE_COMPLETED, TASK_STATE_RETRY, TASK_STATE_SCHEDULED
from django.db.models import Q
from account.models import Member
from datetime import datetime, timedelta
import incident

import os

from rest import settings

logger = getLogger()

PID_FILE = os.path.join(settings.VAR_ROOT, "tq_worker.pid")
ENGINE_CMD = os.path.join(settings.BIN_ROOT, "tq_worker.py")

TQ_SUBSCRIBE = settings.get("TQ_SUBSCRIBE", [])
TQ_REPORT_BACKLOG = settings.get("TQ_REPORT_BACKLOG", 20)
TQ_DELETE_AFTER = settings.get("TQ_DELETE_AFTER", 30)
TQ_DELETE_COMPLETED_AFTER = settings.get("TQ_DELETE_COMPLETED_AFTER", 7)

# allow setting the periodic retry to either 5minutes or 1minute
TQ_RETRY_FREQ = settings.get("TQ_RETRY_FREQ", 5)
if TQ_RETRY_FREQ == 5:
    TQ_RETRY_FREQ = PERIODIC_EVERY_5_MINUTES
else:
    TQ_RETRY_FREQ = None


def isLocalEngineRunning():
    pid = helpers.getPidFromFile(PID_FILE)
    if not pid:
        return False
    return helpers.isPidRunning(pid)


def startLocalEngine():
    # needs to be added to sudoers
    helpers.sudoCMD([ENGINE_CMD, "start"], as_user="www")


@periodic()
def run_check_taskrunner(force=False, verbose=False, now=None):
    # check if pid is running
    # check if any backlog exists (ignore state=1)
    if not TQ_SUBSCRIBE:
        logger.warning("periodic taskrunner called but engine is disabled via empty TQ_SUBSCRIBE")
        return
    if not isLocalEngineRunning():
        # attempt to start local engine
        logger.info("pid is: {} @ {}".format(helpers.getPidFromFile(PID_FILE), PID_FILE))
        logger.info("taskrunner is not running? starting task runner...")
        startLocalEngine()


@periodic(minute=TQ_RETRY_FREQ)
def run_retry(force=False, verbose=False, now=None):
    # check for retry jobs, but only on the master
    if not settings.TQ_MASTER:
        return

    if now is None:
        now = datetime.now()
    retry_jobs = Task.objects.filter(state=TASK_STATE_RETRY).filter(Q(scheduled_for__isnull=True)|Q(scheduled_for__lte=now))[:200]
    for retry in retry_jobs:
        if not retry.is_stale:
            retry.retry_now()
        else:
            retry.failed("stale")


@periodic(minute=PERIODIC_EVERY_5_MINUTES)
def run_checkup(force=False, verbose=False, now=None):
    if not settings.TQ_MASTER:
        return
    # check if we have back log
    qset = Task.objects.filter(state=TASK_STATE_SCHEDULED)
    backlog_count = qset.count()
    # btask = Task.objects.filter(state=TASK_STATE_SCHEDULED).last()
    if backlog_count > TQ_REPORT_BACKLOG:
        details = ""
        for key, value in helpers.countOccurences(qset, "channel").items():
            details += "{} = {}\n".format(key, value)
        incident.event_now(
            "taskqueue_errors", f"Task Queue Backlog of {backlog_count}",
            details=details, level=3)


@periodic(minute=45, hour=10)
def run_cleanup(force=False, verbose=False, now=None):
    if not settings.TQ_MASTER:
        return
    stale = datetime.now() - timedelta(days=TQ_DELETE_AFTER)
    count = Task.objects.filter(created__lte=stale).delete()[0]
    if count:
        logger.info("deleted {} old tasks".format(count))
    stale = datetime.now() - timedelta(days=TQ_DELETE_COMPLETED_AFTER)
    count = Task.objects.filter(created__lte=stale, state=TASK_STATE_COMPLETED).delete()[0]
    if count:
        logger.info("deleted {} old completed tasks".format(count))


PUSH_METRICS_TO_CLOUDWATCH = settings.get("PUSH_METRICS_TO_CLOUDWATCH", False)

@periodic(minute=PERIODIC_EVERY_5_MINUTES)
def run_aws_metrics(force=False, verbose=False, now=None):
    if PUSH_METRICS_TO_CLOUDWATCH:
        from metrics.providers import aws
        aws.publishLocalMetrics()




from django.db import models
from django.forms.widgets import ClearableFileInput
from django.core.files.uploadedfile import UploadedFile
from django.core import validators
from objict import objict
# from .views import chunkUploadedFile
from .crypto import aes

import re
from datetime import datetime, date

try:
    import phonenumbers
except:
    phonenumbers = None

class ResumableFileInput(ClearableFileInput):
    def value_from_datadict(self, data, files, name):
        upload = super(ResumableFileInput, self).value_from_datadict(data, files, name)
        if not upload:
            if name + '.path' in data and name + '.name' in data and name + '.content_type' in data and name + '.size' in data:
                try:
                    f = open(data.get(name + '.path'))
                except OSError:
                    pass
                else:
                    upload = UploadedFile(file=f, name=data.get(name + '.name'), content_type=data.get(name + '.content_type'), size=data.get(name + '.size'))
        if not upload:
            if name + '_sessionid' in data:
                try:
                    upload = chunkUploadedFile(data.get('__request'), data.get(name + '_sessionid'))
                except OSError:
                    pass

        if upload:
            files[name] = upload
        return upload

class DateField(models.DateField):
    def to_python(self, value):
        if value == None or value == '':
            return None
        elif type(value) in (int,float):
            return date.fromtimestamp(value)
        elif type(value) in (str,str) and re.match('^-?[0-9]+$', value):
            try:
                return date.fromtimestamp(int(value))
            except:
                pass
        if type(value) is datetime:
            return value.date()
        return super(models.DateField, self).to_python(value)

class DateTimeField(models.DateTimeField):
    def to_python(self, value):
        if value == None or value == '':
            return None
        elif type(value) in (int,float):
            return datetime.fromtimestamp(value)
        elif type(value) in (str,str) and re.match('^-?[0-9]+$', value):
            try:
                return datetime.fromtimestamp(int(value))
            except:
                pass
        return super(models.DateTimeField, self).to_python(value)

from decimal import Decimal
class CurrencyField(models.DecimalField):
    def __init__(self, *args, **kwargs):
        default_value = kwargs.pop("default", 0.0)
        kwargs["default"] = default_value

        max_digits = kwargs.pop("max_digits", 12)
        kwargs["max_digits"] = max_digits

        decimal_places = kwargs.pop("decimal_places", 2)
        kwargs["decimal_places"] = decimal_places

        super(CurrencyField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection, *args, **kwargs):
        if value is None:
            return value
        if not isinstance(value, Decimal):
            value = Decimal(value)
        return value.quantize(Decimal("0.01"))

    def to_python(self, value):
        if value is None:
            return value
        if not isinstance(value, Decimal):
            value = Decimal(value)
        return value.quantize(Decimal("0.01"))

class FormattedField(models.CharField):
    TITLE = 1
    UPPERCASE = 2
    LOWERCASE = 3
    PHONE = 5

    def __init__(self, *args, **kwargs):
        self.format_kind = kwargs.pop('format', 0)
        max_length = kwargs.pop('max_length', 254)
        kwargs["max_length"] = max_length
        super(FormattedField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection, *args, **kwargs):
        return FormattedField.format(value, self.format_kind)

    def to_python(self, value):
        return FormattedField.format(value, self.format_kind)

    def get_prep_value(self, value):
        value = super(FormattedField, self).get_prep_value(value)
        return FormattedField.format(value, self.format_kind)

    @staticmethod
    def parsePhone(value, region):
        try:
            pp = phonenumbers.parse(value, region)
        except:
            return None
        return pp

    @staticmethod
    def format(value, format=0):
        if not value:
            return value
        if isinstance(value, list):
            # hot fix for some issues being seen
            value = " ".join(value)
        if format == FormattedField.UPPERCASE:
            return value.upper()
        elif format == FormattedField.LOWERCASE:
            return value.lower()
        elif format == FormattedField.TITLE:
            return value.title()
        elif format == FormattedField.PHONE:
            if phonenumbers:
                try:
                    x = phonenumbers.parse(value, "US")
                    if not x:
                        x = phonenumbers.parse(value, None)
                    if not x:
                        return value.replace(' ', '').replace('.', '-').replace('(', '').replace(')', '-')
                    return phonenumbers.format_number(x, phonenumbers.PhoneNumberFormat.INTERNATIONAL)
                except Exception as err:
                    print(("phone format error: '{}' .. error: {}".format(value, str(err))))
            else:
                return value.replace(' ', '').replace('.', '-').replace('(', '').replace(')', '-')
        return value

class CommaSeparatedListField(models.TextField):
    def __init__(self, max_items=None, min_items=None, *args, **kwargs):
        self.max_items, self.min_items = max_items, min_items
        super(CommaSeparatedListField, self).__init__(*args, **kwargs)
        if min_items is not None:
            self.validators.append(validators.MinLengthValidator(min_items))
        if max_items is not None:
            self.validators.append(validators.MaxLengthValidator(max_items))

    def to_python(self, value):
        if value in validators.EMPTY_VALUES:
            return []
        if isinstance(value, str):
            value = value.split(',')
        ret = []
        for t in value:
            if not isinstance(value, str):
                t = str(t)
            t = t.strip()
            if bool(t):
                ret.append(t)
        return ret

    def from_db_value(self, value, expression, connection, *args, **kwargs):
        return self.to_python(value)

    def to_string(self, value):
        if value is None:
            if not self.null and self.blank:
                return ""
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, list):
            return ",".join(value)
        return value

    def get_prep_value(self, value):
        return self.to_string(value)

    def value_to_string(self, obj):
        return self.to_string(obj)


class JSONField(models.TextField):
    def __init__(self, *args, **kwargs):
        default_value = kwargs.pop("default", None)
        null = kwargs.pop("null", True)
        kwargs["default"] = default_value
        kwargs["null"] = null
        super(JSONField, self).__init__(*args, **kwargs)

    # def pre_save(self, model_instance, add):
    #     return super(JSONField, self).pre_save(model_instance, add)

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, str):
            value = objict.fromJSON(value)
        elif isinstance(value, dict) and not hasattr(value, "fromdict"):
            value = objict.fromdict(value)
        return value

    def from_db_value(self, value, expression, connection, *args, **kwargs):
        return self.to_python(value)

    def to_string(self, value):
        if value is None:
            if not self.null and self.blank:
                return ""
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, dict):
            if not hasattr(value, "toJSON"):
                value = objict.fromdict(value)
            return value.toJSON(as_string=True)
        return value

    def get_prep_value(self, value):
        return self.to_string(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        return super().get_db_prep_value(self.to_string(value), connection, prepared)

    # def get_db_prep_save(self, value, connection):
    #     return super().get_db_prep_save(value, connection)

    def value_to_string(self, obj):
        return self.get_prep_value(obj)


class EncryptedField(models.TextField):
    ENCRYPTION_KEY = "&F)J@NcRfUjXn2r5u8x/A?D*G-KaPdSgVkYp3s6v9y$B&E)H+MbQeThWmZq4t7w!z%C*F-JaNcRfUjXn2r5u8x/A?D(G+KbPeSgVkYp3s6v9y$B&E)H@McQfTjWmZq4t"

    def __init__(self, *args, **kwargs):
        default_value = kwargs.pop("default", None)
        null = kwargs.pop("null", True)
        kwargs["default"] = default_value
        kwargs["null"] = null
        super(EncryptedField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection, *args, **kwargs):
        if value is None:
            return value
        return EncryptedField.decrypt(value)

    def to_python(self, value):
        if value is None:
            return value
        return EncryptedField.decrypt(value)

    def get_prep_value(self, value):
        if value is None:
            if not self.null and self.blank:
                return ""
            return None
        return EncryptedField.encrypt(value)

    @staticmethod
    def encrypt(text):
    	return aes.encrypt(EncryptedField.ENCRYPTION_KEY, text)

    @staticmethod
    def decrypt(encrypted):
        return aes.decrypt(EncryptedField.ENCRYPTION_KEY, encrypted)


import requests
import ipaddress
from objict import objict
from rest import settings
from datetime import datetime
from enum import Enum

API_KEY = settings.get("ABUSEIPDB_KEY", "")

DEFAULT_HEADERS = {
    'Accept': 'application/json',
    'Key': API_KEY
}


class AbuseCategory(Enum):
    DNS_COMPROMISE = (1, "Altering DNS records resulting in improper redirection.")
    DNS_POISONING = (2, "Falsifying domain server cache (cache poisoning).")
    FRAUD_ORDERS = (3, "Fraudulent orders.")
    DDOS_ATTACK = (4, "Participating in distributed denial-of-service (usually part of botnet).")
    FTP_BRUTE_FORCE = (5, "FTP Brute-Force")
    PING_OF_DEATH = (6, "Oversized IP packet.")
    PHISHING = (7, "Phishing websites and/or email.")
    FRAUD_VOIP = (8, "Fraud VoIP")
    OPEN_PROXY = (9, "Open proxy, open relay, or Tor exit node.")
    WEB_SPAM = (10, "Comment/forum spam, HTTP referer spam, or other CMS spam.")
    EMAIL_SPAM = (11, "Spam email content, infected attachments, and phishing emails.")
    BLOG_SPAM = (12, "CMS blog comment spam.")
    VPN_IP = (13, "Conjunctive category.")
    PORT_SCAN = (14, "Scanning for open ports and vulnerable services.")
    HACKING = (15, "Hacking")
    SQL_INJECTION = (16, "Attempts at SQL injection.")
    SPOOFING = (17, "Email sender spoofing.")
    BRUTE_FORCE = (18, "Credential brute-force attacks on webpage logins and services.")
    BAD_WEB_BOT = (19, "Webpage scraping and crawlers that do not honor robots.txt.")
    EXPLOITED_HOST = (20, "Host is likely infected with malware and being used for other attacks.")
    WEB_APP_ATTACK = (21, "Attempts to probe for or exploit installed web applications.")
    SSH = (22, "Secure Shell (SSH) abuse.")
    IOT_TARGETED = (23, "Abuse targeted at an 'Internet of Things' type device.")

    def __init__(self, code, description):
        self.code = code
        self.description = description

    def __str__(self):
        return f"{self.name} ({self.code}): {self.description}"


def lookup(ip, max_days=0):
    """
    {
      "data": {
        "ipAddress": "118.25.6.39",
        "isPublic": true,
        "ipVersion": 4,
        "isWhitelisted": false,
        "abuseConfidenceScore": 100,
        "countryCode": "CN",
        "countryName": "China",
        "usageType": "Data Center/Web Hosting/Transit",
        "isp": "Tencent Cloud Computing (Beijing) Co. Ltd",
        "domain": "tencent.com",
        "hostnames": [],
        "isTor": false,
        "totalReports": 1,
        "numDistinctUsers": 1,
        "lastReportedAt": "2018-12-20T20:55:14+00:00",
        "reports": [
          {
            "reportedAt": "2018-12-20T20:55:14+00:00",
            "comment": "Dec 20 20:55:14 srv206 sshd[13937]: Invalid user oracle from 118.25.6.39",
            "categories": [
              18,
              22
            ],
            "reporterId": 1,
            "reporterCountryCode": "US",
            "reporterCountryName": "United States"
          }
        ]
      }
    }
    """
    if not API_KEY:
        return objict(errors=[{"detail": "Missing API KEY", "status": 401}])
    url = 'https://api.abuseipdb.com/api/v2/check'
    params = dict(ipAddress=ip)
    if max_days:
        params["maxAgeInDays"] = max_days
    response = requests.request(
        method='GET', url=url, headers=DEFAULT_HEADERS,
        params=params)
    output = objict.fromJSON(response.text)
    if "data" in output:
        return output.data
    return output


def reportAbuse(ip, comment, categories):
    """
    {
        "data": {
            "ipAddress": "127.0.0.1",
            "abuseConfidenceScore": 52
        }
    }
    """
    if not API_KEY:
        return objict(errors=[{"detail": "Missing API KEY", "status": 401}])
    url = 'https://api.abuseipdb.com/api/v2/report'
    now_str = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    params = dict(ip=ip, comment=comment, timestamp=now_str, categories=categories)
    response = requests.request(
        method='POST', url=url,
        headers=DEFAULT_HEADERS, params=params)
    return objict.fromJSON(response.text)


def isPrivate(ip, ignore_errors=False):
    if not ignore_errors:
        ip_obj = ipaddress.ip_address(ip)
        return ip_obj.is_private
    try:
        ip_obj = ipaddress.ip_address(ip)
        return ip_obj.is_private
    except Exception:
        pass
    return None



import boto3
from datetime import datetime, timedelta
import re
import time
import json
from objict import objict
from rest import settings
from rest import datem
from rest import helpers as rh
from concurrent.futures import ThreadPoolExecutor


LOG_CACHE = objict()


def getClient():
    if LOG_CACHE.client is None:
        key = settings.AWS_KEY
        secret = settings.AWS_SECRET
        region = settings.AWS_REGION
        LOG_CACHE.client = boto3.client("logs", aws_access_key_id=key, aws_secret_access_key=secret, region_name=region)
    return LOG_CACHE.client


def log(data, log_group, log_stream):
    if LOG_CACHE.pool is None:
        LOG_CACHE.pool = ThreadPoolExecutor(max_workers=1)
    LOG_CACHE.pool.submit(logToCloudWatch, data, log_group, log_stream)
    return True


def get(log_group, log_stream, size=20):
    client = getClient()
    if log_stream is None:
        log_stream = DEFAULT_LOG_STREAMS
    if isinstance(log_stream, list):
        log_events = client.get_log_events(
            logGroupName=log_group,
            logStreamNames=log_stream,
            limit=size,
            startFromHead=False
        )
    else:
        log_events = client.get_log_events(
            logGroupName=log_group,
            logStreamName=log_stream,
            limit=size,
            startFromHead=False
        )

    return log_events



def logToCloudWatch(message, log_group, log_stream):
    if isinstance(message, dict):
        message = json.dumps(message)
    return logBatchToCloudWatch([
            dict(
                timestamp=int(datetime.utcnow().timestamp() * 1000),
                message=message)
        ], log_group, log_stream)


def logBatchToCloudWatch(batch, log_group, log_stream):
    return getClient().put_log_events(
        logGroupName=log_group,
        logStreamName=log_stream,
        logEvents=batch
    )


def getLogGroups(names_only=True):
    response = getClient().describe_log_groups()
    groups = response.get('logGroups', [])
    if names_only:
        return [item["logGroupName"] for item in groups]
    return groups


def getLogStreams(log_group):
    log_streams = client.describe_log_streams(
        logGroupName=log_group_name
    )
    return log_streams['logStreams']


def createLogStream(log_group, log_stream):
    try:
        getClient().create_log_stream(logGroupName=log_group, logStreamName=log_stream)
    except Exception:
        pass  # Log stream already exists, no need to create it


def filter(log_group, log_streams, pattern, start_time="1h", end_time=None, resp_format=None):
    if log_streams is None:
        log_streams = DEFAULT_LOG_STREAMS

    start_time, end_time = datem.convert_to_epoch_range(start_time, end_time)
    rh.debug("filter", start_time, end_time, pattern, log_group, log_streams)
    client = getClient()
    response =client.filter_log_events(
        logGroupName=log_group,
        logStreamNames=log_streams,
        filterPattern=f'"{pattern}"',
        startTime=start_time,
        endTime=end_time,
        limit=1000
    )
    rh.debug("filter_log_events", response)
    out = []
    out.extend(response["events"])
    prev_next = None
    while "nextToken" in response and response["nextToken"]:
        if prev_next == response["nextToken"]:
            break
        prev_next = response["nextToken"] 
        response = client.filter_log_events(
            logGroupName=log_group,
            logStreamNames=log_streams,
            filterPattern=f'"{pattern}"',
            startTime=start_time,
            endTime=end_time,
            nextToken=response["nextToken"]
        )
        rh.debug("filter_log_events", response)
        out.extend(response["events"])
    if resp_format == "nginx":
        return dict(events=parseNginxEvents(out))
    return dict(events=out)


QUERY_BY_IP = """fields @message 
| filter @logStream in {log_streams}
| filter @message like /^{ip}/"""

QUERY_BY_TEXT = """fields @message 
| filter @logStream in {log_streams}
| filter @message like "{text}" """

DEFAULT_LOG_STREAMS = ["access.log", "ui_access.log"]

def startSearch(log_groups, start_time, end_time=None, 
                text=None, ip=None, query_string=None,
                log_streams=DEFAULT_LOG_STREAMS):
    if log_streams is None:
        log_streams = DEFAULT_LOG_STREAMS
    # 173\.196\.133\.90
    if ip is not None and "." in ip:
        query_string = QUERY_BY_IP.format(ip=ip.replace('.', '\.'), log_streams=log_streams)
    elif text is not None:
        query_string = QUERY_BY_TEXT.format(text=text, log_streams=log_streams)
    return startInsights(log_groups, start_time, end_time, query_string)


def startInsights(log_groups, start_time, end_time, query_string):
    """
    Executes a CloudWatch Logs Insights query and returns the results.

    :param log_group: The name of the log group to query.
    :param start_time: The start time of the query (epoch time in seconds).
    :param end_time: The end time of the query (epoch time in seconds).
    :param query_string: The query string to use.
    :param region_name: AWS region name.
    :return: The query results.
    """
    # Create a CloudWatch Logs client
    client = getClient()
    start_time, end_time = datem.convert_to_epoch_range(start_time, end_time)

    if log_groups is None:
        log_groups = getLogGroups()
    
    # Start the query
    start_query_response = client.start_query(
        logGroupNames=log_groups,
        startTime=start_time,
        endTime=end_time,
        queryString=query_string,
    )
    return dict(query_id=start_query_response['queryId'], query=query_string)
    

def getInsightResults(query_id, resp_format=None):
    # Wait for the query to complete
    resp = getClient().get_query_results(queryId=query_id)
    status = resp["status"].lower()
    results = insightsToMessage(resp["results"])
    stats = resp["statistics"]
    if status == "complete" and resp_format == "nginx":
        results = parseNginxEvents(results)
    return dict(status=True, state=status, stats=stats, data=results, count=len(results))


def insightsToMessage(events):
    out = []
    for fields in events:
        obj = {}
        for field in fields:
            if "field" in field and field["field"][0] == "@":
                obj[field["field"][1:]] = field["value"]
        if obj:
            out.append(obj)
    return out

def parseNginxEvents(events):
    pattern = re.compile(
        r'(?P<ip>\d+\.\d+\.\d+\.\d+) - - \[(?P<time>.+?)\] '
        r'"(?P<method>\w+) (?P<url>.+?) (?P<protocol>[\w/.]+)" '
        r'(?P<status>\d+) (?P<bytes>\d+) "(?P<referer>.+?)" '
        r'"(?P<user_agent>.+?)" (?P<request_time>\S+) (?P<server_port>\d+)'
    )
    return [pattern.match(line["message"]).groupdict() for line in events]



# Generated by Django 4.1.4 on 2023-11-27 22:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('telephony', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sms',
            name='sid',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=125, null=True),
        ),
    ]



from datetime import datetime, timedelta
from objict import objict
from account.models import Group
from rest import log
import random

from metrics import eod as em

keys = ["count_total", "count_paid", "count_declined", "amount_cents", "fees_cents"]
slug = "testing.set1"
group = Group.objects.last()
# end of day will be 2:30 Central Time
tz = "US/Central"
eod_hour = 14
eod_minute = 30

# for testing history
start = datetime.now() - timedelta(hours=60)

for i in range(0, 60):
    # this will keep track of running sums
    # each time it is called all the fields will be added to the existing values 
    start = start + timedelta(hours=1)
    data = dict(count_paid=0, count_declined=0, amount_cents=0, fees_cents=0)
    data["count_total"] = 1
    if random.randint(0,10) > 7:
        data["count_paid"] = 1
        data["amount_cents"] = random.randint(5, 300) * 100
        data["fees_cents"] = random.randint(1, 8) * 100
    # when is not needed but for the sake of testing we will pass in a when
    # first lets call without a group to get the all
    em.metric(keys, slug, data, when=start, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute)
    # now we can get these metrics
    log.pp(start, data, em.get_metric(slug, start, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute))
    # next lets call with a group to add group specific
    em.metric(keys, slug, data, when=start, group=group, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute)
    log.pp(
        start, data,
        em.get_metric(slug, start, group=group, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute))



log.pp(em.get_metrics(slug, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute))



# Generated by Django 4.2.11 on 2024-05-13 02:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0014_event_group_alter_rulecheck_index'),
    ]

    operations = [
        migrations.AddField(
            model_name='rule',
            name='title_template',
            field=models.CharField(default=None, max_length=200, null=True),
        ),
        migrations.AlterField(
            model_name='incident',
            name='state',
            field=models.IntegerField(choices=[(0, 'new'), (1, 'opened'), (2, 'paused'), (3, 'ignored'), (4, 'resolved'), (5, 'pending')], default=0),
        ),
    ]



from __future__ import unicode_literals
from django.core.management.base import BaseCommand, CommandError
from optparse import make_option
from redis_metrics.utils import generate_test_metrics


class Command(BaseCommand):
    args = '<metric-name> [<metric-name> ...]'
    help = "Creates Lots of Dummy Metrics"
    option_list = BaseCommand.option_list + (
        make_option(
            '-r',
            '--randomize',
            action='store_true',
            dest='randomize',
            default=True,
            help='Randomize Metric Data'
        ),
        make_option(
            '--no-randomize',
            action='store_false',
            dest='randomize',
            default=True,
            help='Do not randomize Metric Data'
        ),
        make_option(
            '-n',
            '--num-days',
            action='store',
            dest='num_days',
            type="int",
            default=365 * 3,  # Default to 3 years
            help='Number of Days worth of data to generate'
        ),
        make_option(
            '-c',
            '--cap',
            action='store',
            dest='cap',
            default=None,
            help='Cap the maximum metric value'
        ),
    )

    def handle(self, *args, **options):
        if len(args) == 0:
            raise CommandError("You must provide at least one metric name")

        slugs = args
        cap = options["cap"]
        days = options["num_days"]
        randomize = options["randomize"]

        self.stdout.write("\nGenerating metrics using the following:\n")
        self.stdout.write("Slugs: {0}\n".format(u", ".join(slugs)))
        self.stdout.write("Days: {0}\n".format(days))
        self.stdout.write("Randomize: {0}\n".format(randomize))
        self.stdout.write("Cap: {0}\n".format(cap))

        for slug in slugs:
            generate_test_metrics(slug, num=days, randomize=randomize, cap=cap)



# Generated by Django 4.2.11 on 2024-05-13 02:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sessionlog', '0001_initial'),
        ('auditlog', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='persistentlog',
            name='session',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='sessionlog.sessionlog'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-05-13 02:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('inbox', '0004_mailtemplate'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mailbox',
            name='state',
            field=models.IntegerField(db_index=True, default=1),
        ),
    ]
